## file for florence functions

import torch
from PIL import Image, ImageDraw
from transformers import AutoProcessor, AutoModelForCausalLM
from matplotlib import pyplot as plt
import matplotlib.patches as patches
import numpy as np
import pandas as pd
from transformers import (
    AdamW,
    AutoModelForCausalLM,
    AutoProcessor,
    get_scheduler
)
from tqdm import tqdm
from typing import List, Dict, Any, Tuple, Generator
from peft import LoraConfig, get_peft_model
from torch.utils.data import Dataset, DataLoader
import os

__all__ = ["florence_model","load_florence_dataset","dataset_data"]

class florence_model:
    """
    Class for florence model
    Args:
        model_name (str): Name of the model
        device (str, optional): Device on which to run the model. Defaults to 'cpu'.
    Returns:
        None
    """

    def __init__(self,model_name="microsoft/Florence-2-large",device='cpu') -> None:

        if device == 'cpu':
            device = "cpu"
        elif device == 'cuda':
            device = "cuda:0" if torch.cuda.is_available() else "cpu"
        else:
            device = device
        self.device = device
        self.model_name = model_name
        self.model = AutoModelForCausalLM.from_pretrained(model_name,trust_remote_code=True).to(device)
        self.processor = AutoProcessor.from_pretrained(model_name,trust_remote_code=True)

    def get_task_types(self):
        """
        Function to get task types
        Args:
            None
        Returns:
            task_type (dict): List of task types
        """
        task_list =  {'captions':['<CAPTION>','<DETAILED_CAPTION>','<MORE_DETAILED_CAPTION>'],'character_recognition':['<OCR>','<OCR_WITH_REGION>'],
                      'object_detection':['<OD>','<REGION_PROPOSAL>','<DENSE_REGION_PROPOSAL>'],'segmentation':['<REGION_TO_SEGMENTATION>'],
                      'description':['<REGION_TO_CATOGORY>','<REGION_TO_DESCRIPTION>'],
                      'extra':['<PHRASE_GROUNDING>','<OPEN_VOCABULARY_DETECTION>','<REFERRING_EXPRESSION_SEGMENTATION>']}
        return task_list
    
    def define_task(self,task_type:list = ['<OD>','CAPTION']):
        """
        Function to define task
        Args:
            task_type (list): List of task types
        Returns:
            None
        """
        self.task_type = task_type

    def set_image(self,image_path:str):
        """
        Function to set image
        Args:
            image_path (str): Image path
        Returns:
            None
        """
        self.image = Image.open(image_path)

    def generate_text(self,prompt:str =None):
        """
        Function to generate text
        Args:
            prompt (str): Prompt
        Returns:
            res (str): Result
        """
        final_ans = []
        for i in self.task_type:
            if prompt:
                prompt = i + prompt
            else:
                prompt = i
        
            inputs = self.processor(text=prompt, images=self.image, return_tensors="pt").to(self.device)
            output_ids = self.model.generate(inputs["input_ids"],pixel_values=inputs["pixel_values"],
                                             max_new_tokens=1024,early_stopping=False,do_sample=False,num_beams=3)
            generated_text = self.processor.batch_decode(output_ids, skip_special_tokens=False)[0]
            parsed_answer = self.processor.post_process_generation(generated_text, task=i, 
                                                                image_size=(self.image.width, self.image.height))
            final_ans.append(parsed_answer)
        return final_ans
    
    def plot_box(self,data,image=None,show=True,save_path=None):
        """
        Plot BBox on image
        Args:
            data (dict): BBox data
            image (PIL Image): Image
        Returns:
            fig (matplotlib figure): Figure
        """
        fig, ax = plt.subplots(figsize=(12, 8))
        if image is None:
            image = self.image
        if show:
            ax.imshow(image)

        for bbox, label in zip(data['bboxes'], data['labels']):
            x1, y1, x2, y2 = bbox
            rect = patches.Rectangle((x1, y1),
                                 x2 - x1,
                                 y2 - y1,
                                 linewidth=2,
                                 edgecolor='lime',
                                 facecolor='none')
            ax.add_patch(rect)
            plt.text(x1,
                 y1,
                 label,
                 color='black',
                 fontsize=8,
                 bbox=dict(facecolor='lime', alpha=1))

        ax.axis('off')
        if show:
            plt.show()
        if save_path is not None:
            fig.savefig(save_path)

    def plot_bboxes_on_image(self,image, bboxes, labels):
        plt.figure(figsize=(10, 8))
        plt.imshow(image)

        ax = plt.gca()
        for bbox, label in zip(bboxes, labels):
            xmin, ymin, xmax, ymax = bbox
            width = xmax - xmin
            height = ymax - ymin
            rect = plt.Rectangle((xmin, ymin), width, height, fill=False, edgecolor='red', linewidth=2)
            ax.add_patch(rect)
            ax.text(xmin, ymin - 2, label, bbox=dict(facecolor='red', alpha=0.5), fontsize=12, color='white')
        plt.axis('off')
        plt.show()

    def draw_polygons(self, prediction,image=None, fill_mask=False,show=True,save_path=None):
        """
        Draws segmentation masks with polygons on an image.
        Args:
            image (PIL Image): The image to draw on.
            prediction (dict): The prediction containing polygons and labels.
            fill_mask (bool): Whether to fill the polygons with color.
        """
        if image is None:
            image = self.image
        draw = ImageDraw.Draw(image)
        scale = 1

        for polygons, label in zip(prediction['polygons'], prediction['labels']):
            color = "lime"
            fill_color = "lime" if fill_mask else None

            for _polygon in polygons:
                _polygon = np.array(_polygon).reshape(-1, 2)
                if len(_polygon) < 3:
                    print('Invalid polygon:', _polygon)
                    continue

                _polygon = (_polygon * scale).reshape(-1).tolist()
                if fill_mask:
                    draw.polygon(_polygon, outline=color, fill=fill_color)
                else:
                    draw.polygon(_polygon, outline=color)
                draw.text((_polygon[0] + 8, _polygon[1] + 2), label, fill=color)
        if show:
            plt.imshow(image)
        if save_path is not None:
            image.save(save_path)

        return image

    def convert_to_od_format(self,data):
        """
        Convert data to od format
        Args:
            data (dict): Data
        Returns:
            od_results (dict): od results
        """
        bboxes = data.get('bboxes', [])
        labels = data.get('bboxes_labels', [])
        od_results = {'bboxes': bboxes, 'labels': labels}

        return od_results
    
    def draw_ocr_bbox(self,data,image=None,show=True,save_path=None):
        """
        Draw OCR BBox
        Args:
            data (dict): Data
            image (PIL Image): Image
            show (bool): Show
            save_path (str): Save path
        Returns:
            fig (matplotlib figure): Figure
        """
        scale = 1
        if image is None:
            image = self.image
        
        draw = ImageDraw.Draw(image)
        bboxes,labels= data['bboxes'],data['labels']

        for bbox, label in zip(bboxes, labels):
            color='lime'
            new_box= (np.array(bbox)*scale).tolist()
            draw.polygon(new_box,width=4, outline=color)
            draw.text((new_box[0]+8, new_box[0][1]+2), "{}".format(label),align='right', fill=color)

        if show:
            plt.imshow(image)
        if save_path is not None:
            image.save(save_path)
        return image
    
    def load_model(self,model_path:str):
        """
        Function to load model
        Args:
            model_path (str): Model path
        Returns:
            None
        """
        self.model = AutoModelForCausalLM.from_pretrained(model_path).to(self.device)

    def my_collate_fn(self,batch):
        """
        Collate function for the dataloader
        Args:
            batch (list): List of data
        Returns:
            dict: Dictionary of data
        """
        prefix = [item[0] for item in batch]
        suffix = [item[1] for item in batch]
        image = [item[2] for item in batch]
        inputs = self.processor(text=list(prefix), images=list(image), return_tensors="pt", padding=True).to('cpu')
        return inputs, suffix
    
    def dataloader(self,dataset:Dataset,batch_size:int):
        """
        Function to load dataloader
        Args:
            dataset (Dataset): Dataset
            batch_size (int): Batch size
        Returns:
            dataloader
        """
        dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=self.my_collate_fn,shuffle=True)
        return dataloader

    def load_lora_data(self):
        """
        Function to load lora data.
        Default param set. Can be changed as per requirement. Currently hard coded
        Args:
            None
        Returns:
            peft_model
        """
        config = LoraConfig(r=8,lora_alpha=8,
                            target_modules=["q_proj", "o_proj", "k_proj", "v_proj", "linear", "Conv2d", "lm_head", "fc2"],
                            task_type="CAUSAL_LM",
                            lora_dropout=0.05,
                            bias="none",
                            inference_mode=False,
                            use_rslora=True,
                            init_lora_weights="gaussian",)

        self.peft_model = get_peft_model(self.model, config)
        self.peft_model.print_trainable_parameters()

        return self.peft_model
    
    def florence2_inference_results(self, dataset: Dataset, count: int):
        count = min(count, len(dataset.dataset))
        for i in range(count):
            prefix,suffix, image = dataset.dataset[i]
            inputs = self.processor(text=prefix, images=image, return_tensors="pt").to(self.device)
        #inputs = processor(text=prompt, images=image, return_tensors="pt")
    
            generated_ids = self.peft_model.generate(
                input_ids=inputs["input_ids"],
                pixel_values=inputs["pixel_values"],
                max_new_tokens=1024,
                early_stopping=False,
                do_sample=False,
                num_beams=3,)
            
            generated_text = self.processor.batch_decode(generated_ids,
                                                skip_special_tokens=False)[0]
            parsed_answer = self.processor.post_process_generation(generated_text,task='<OD>',image_size=(image.width, image.height))
            # Access bounding boxes and labels
            od_results = parsed_answer['<OD>']
            bboxes = od_results['bboxes']
            labels = od_results['labels']
        
            # Plot bounding boxes on the image using the separate function
            self.plot_bboxes_on_image(image, bboxes, labels)
        return parsed_answer
    
    def train_model(self,train_loader, val_loader, epochs=10, lr=1e-6):
        optimizer = AdamW(self.peft_model.parameters(), lr=lr)
        num_training_steps = epochs * len(train_loader)
        lr_scheduler = get_scheduler(
            name="linear",
            optimizer=optimizer,
            num_warmup_steps=0,
            num_training_steps=num_training_steps,)

        #florence2_inference_results(peft_model, val_loader.dataset, 3)

        for epoch in range(epochs):
            self.peft_model.train()
            train_loss = 0
            for inputs, answers in tqdm(train_loader, desc=f"Training Epoch {epoch + 1}/{epochs}"):

                input_ids = inputs["input_ids"]
                pixel_values = inputs["pixel_values"]
                labels = self.processor.tokenizer(
                    text=answers,
                    return_tensors="pt",
                    padding=True,
                    return_token_type_ids=False).input_ids.to(self.device)

                outputs = self.peft_model(input_ids=input_ids, pixel_values=pixel_values, labels=labels)
                loss = outputs.loss

                loss.backward(), optimizer.step(), lr_scheduler.step(), optimizer.zero_grad()
                train_loss += loss.item()

            avg_train_loss = train_loss / len(train_loader)
            print(f"Average Training Loss: {avg_train_loss}")

            self.peft_model.eval()
            val_loss = 0
            with torch.no_grad():
                for inputs, answers in tqdm(val_loader, desc=f"Validation Epoch {epoch + 1}/{epochs}"):

                    input_ids = inputs["input_ids"]
                    pixel_values = inputs["pixel_values"]
                    labels = self.processor.tokenizer(
                        text=answers,
                        return_tensors="pt",
                        padding=True,
                        return_token_type_ids=False).input_ids.to(self.device)

                    outputs = self.peft_model(input_ids=input_ids, pixel_values=pixel_values, labels=labels)
                    loss = outputs.loss

                    val_loss += loss.item()

                avg_val_loss = val_loss / len(val_loader)
                print(f"Average Validation Loss: {avg_val_loss}")

                self.florence2_inference_results(self.peft_model, val_loader.dataset, 2)

            output_dir = f"./model_checkpoints/epoch_{epoch+1}"
            os.makedirs(output_dir, exist_ok=True)
            self.model.save_pretrained(output_dir)
            self.processor.save_pretrained(output_dir)

class load_florence_dataset:
    """
    Requires csv file path.
    CSV file should have:
        image path (image_path) ,
        boundingbox or suffix (bbox in the format [xmin, ymin, xmax, ymax] -should be in order of labels.) -list[list], 
        prefix (if not mentioned, default <OD> will be used), 
        labels or suffix (labels will be used as suffix after converting. if suffix is used, please make it in proper format) -list[list]
        train_type : train or validation (if not mentioned, default random 80:20 will be used)
        
    Check the test file for more details. (./test_data/florence_test.csv)   
    Args:
        csv_file_path (str): Path to the csv file
    Returns:
        None
    """

    def __init__(self,csv_file_path) -> None:
        self.df = pd.read_csv(csv_file_path)
        if 'bbox' not in self.df.columns and 'suffix' not in self.df.columns:
            raise ValueError("CSV file should have 'bbox' column or 'suffix' columns")
        if 'labels' not in self.df.columns and 'suffix' not in self.df.columns:
            raise ValueError("CSV file should have 'labels' or 'suffix' columns")
        if 'prefix' not in self.df.columns:
            self.df['prefix'] = '<OD>'
        if 'prefix' in self.df.columns:
            prefix_type_list =  ['<CAPTION>','<DETAILED_CAPTION>','<MORE_DETAILED_CAPTION>','<OCR>','<OCR_WITH_REGION>',
                                 '<OD>','<REGION_PROPOSAL>','<DENSE_REGION_PROPOSAL>','<REGION_TO_SEGMENTATION>',
                                 '<REGION_TO_CATOGORY>','<REGION_TO_DESCRIPTION>',
                                 '<PHRASE_GROUNDING>','<OPEN_VOCABULARY_DETECTION>','<REFERRING_EXPRESSION_SEGMENTATION>']
            prefix_u = list(self.df['prefix'].unique())
            assert len(prefix_u)==1,'Check prefix - multiple values'
            assert type(prefix_type_list.index(prefix_u[0]))==int,'Check prefix value. Not in index'
        
        if 'train_type' not in self.df.columns:
            self.df['train_type'] = 'train'
            val_indices = np.random.choice(self.df.index, size=int(len(self.df) * 0.2), replace=False)
            self.df.loc[val_indices, 'train_type'] = 'validation'

        final_data = []
        if 'suffix' not in self.df.columns:
            self.df['suffix'] = None
            temp_dict={}
            prefix_val = self.df.prefix.iloc[0]
            for i, row in self.df.iterrows():
                if isinstance(row['bbox'], str):
                    bbox_list = eval(row['bbox'])
                else:
                    bbox_list = row['bbox']
                    
                if isinstance(row['labels'], str):
                    try:
                        labels_list = eval(row['labels'])
                    except:
                        labels_list = [row['labels']]
                else:
                    labels_list = row['labels']
                assert len(bbox_list)==len(labels_list),f'BBox list and labels list not equal in row {i}'
                temp_str = ""
                for j in range(len(bbox_list)):
                    if isinstance(bbox_list[j],str):
                        bbox_list[j] = eval(bbox_list[j])
                    temp_str = temp_str+ f"{labels_list[j]}<loc_{int(bbox_list[j][0])}><loc_{int(bbox_list[j][1])}><loc_{int(bbox_list[j][2])}><loc_{int(bbox_list[j][3])}>"
                self.df.at[i, 'suffix'] = temp_str
                temp_dict['image'] = row['image_path']
                temp_dict['prefix'] = prefix_val
                temp_dict['suffix'] = temp_str
                final_data.append(temp_dict)
        self.final_csv = pd.DataFrame(final_data)
        self.final_csv.to_csv('./final_data_florence.csv',index=False)
        print(f'final data example : {final_data[0]}')
        print(f'final csv example: {self.final_csv.head(2)}')
            
    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        return self.final_csv.iloc[idx]

class dataset_data(Dataset):
    def __init__(self,df) -> None:
        self.df = df
    
    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        # row_data = self.final_csv[self.final_csv.index==idx] ## check how to seperate train/val before goin to this step
        image_path = self.df.iloc[idx]['image_path']
        prefix = self.df.iloc[idx]['prefix']
        suffix = self.df.iloc[idx]['suffix']
        try:
            image = Image.open(image_path)
        except:
            print(f"Error opening image: {image_path}")
        return prefix, suffix, image
    




# def collate_fn(batch):
#     """
#     Collate function for the dataloader
#     Args:
#         batch (list): List of data
#     Returns:
#         dict: Dictionary of data
#     """
#     prefix = [item[0] for item in batch]
#     suffix = [item[1] for item in batch]
#     image = [item[2] for item in batch]
#     inputs = processor(text=list(prefix), images=list(image), return_tensors="pt", padding=True).to('cpu')
#     return inputs, suffix




## function for segment anything 2

import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from sam2.build_sam import build_sam2
from sam2.automatic_mask_generator import SAM2AutomaticMaskGenerator
import cv2
from sam2.build_sam import build_sam2_video_predictor
from os import listdir
from os.path import isfile, join
from sam2.sam2_image_predictor import SAM2ImagePredictor


__all__ = ["show_anns","get_mask_generator","get_mask_for_bbox","get_all_masks","video_predictor",
           "show_masks_image","show_box","show_points","image_predictor"]

def show_anns(anns, borders=True, show=True):
    """
    show the annotations
    Args:
        anns (list): list of annotations
        borders (bool): if True, show the borders of the annotations
    Returns:
        None
    """
    if len(anns) == 0:
        return
    sorted_anns = sorted(anns, key=(lambda x: x['area']), reverse=True)
    ax = plt.gca()
    ax.set_autoscale_on(False)
 
    img = np.ones((sorted_anns[0]['segmentation'].shape[0], sorted_anns[0]['segmentation'].shape[1], 4))
    img[:,:,3] = 0
    for ann in sorted_anns:
        m = ann['segmentation']
        color_mask = np.concatenate([np.random.random(3), [0.5]])
        img[m] = color_mask 
        if borders:
            import cv2
            contours, _ = cv2.findContours(m.astype(np.uint8),cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE) 
            # Try to smooth contours
            contours = [cv2.approxPolyDP(contour, epsilon=0.01, closed=True) for contour in contours]
            cv2.drawContours(img, contours, -1, (0,0,1,0.4), thickness=1) 
    if show:    
        ax.imshow(img)
    return img


def get_mask_generator(sam2_checkpoint='../checkpoints/sam2_hiera_large.pt', model_cfg='sam2_hiera_l.yaml',device='cpu'):
    """
    get the mask generator
    Args:
        sam2_checkpoint (str): path to the sam2 checkpoint
        model_cfg (str): path to the model configuration
    Returns:
        mask_generator (SAM2AutomaticMaskGenerator): mask generator
    """

    sam2 = build_sam2(model_cfg, sam2_checkpoint, device =device, apply_postprocessing=False)
    mask_generator = SAM2AutomaticMaskGenerator(sam2)
    return mask_generator

def get_similarity_value(box1,box2):
    val1 = abs(box1[0]-box2[0])
    val2 = abs(box1[1]-box2[1])
    val3 = abs(box1[2]-box2[2])
    val4 = abs(box1[3]-box2[3])
    total_val = val1+val2+val3+val4
    return total_val

def get_final_similar_box(box1,box2: list):
    best_box = None
    best_val = None
    index = None
    for i in box2:
        val = get_similarity_value(box1,i)
        if best_box is None or val < best_val:
            best_box = i
            best_val = val
            index = box2.index(i)
    return best_box,index

def get_mask_for_bbox(image_path,bbox_value,sam2_checkpoint,model_cfg,device='cpu',show_full: bool=False,show_final: bool=False,*kwargs):
    """
    get the mask
    Args:
        image_path (str): path to the image
        bbox_value : bounding box value to get mask for
        sam2_checkpoint (str): path to the sam2 checkpoint
        model_cfg (str): path to the model configuration
        show_full (bool): if True, show the full mask
        show_final (bool): if True, show the final mask 
        **kwargs: additional arguments
    Returns:
        mask (np.array): mask
        bbox (list): bounding box
        main_bbox (list): main bounding box
    """
    print('Getting mask')
    image = cv2.imread(image_path)
    image = cv2.cvtColor(image,cv2.COLOR_BGR2RGB)
    mask_generator = get_mask_generator(sam2_checkpoint,model_cfg,device)
    mask_full = mask_generator.generate(image)
    if show_full:
        show_anns(mask_full)
    print('Getting final mask')

    main_bbox = []
    for i in mask_full:
        mask_val = [i['bbox'][1],i['bbox'][0],
                    (i['bbox'][3]+i['bbox'][1]),(i['bbox'][2]+i['bbox'][0])]  ## adding the bbox values to get the correct bbox for the bounding bbox function
        main_bbox.append(mask_val)


    value_list,index = get_final_similar_box(bbox_value,main_bbox)
    final_mask = mask_full[index]
    final_bbox = [final_mask['bbox'][1],final_mask['bbox'][0],
                    (final_mask['bbox'][3]+final_mask['bbox'][1]),(final_mask['bbox'][2]+final_mask['bbox'][0])]
    if show_final:
        show_anns(final_mask)
    return final_mask['segmentation'],final_bbox,main_bbox

def get_all_masks(image_path,sam2_checkpoint,model_cfg,device='cpu',*kwargs):
    """
    get all the masks
    Args:
        image_path (str): path to the image
        sam2_checkpoint (str): path to the sam2 checkpoint
        model_cfg (str): path to the model configuration
    Returns:
        masks (list): list of masks
    """
    print('Getting all masks')
    image = cv2.imread(image_path)
    image = cv2.cvtColor(image,cv2.COLOR_BGR2RGB)
    mask_generator = get_mask_generator(sam2_checkpoint,model_cfg,device)
    mask_full = mask_generator.generate(image)
    print('Getting final mask')
    return mask_full


def show_mask(mask, ax, obj_id=None, random_color=False):
    """
    Display a mask on a given axis.
    Args:
        mask (np.ndarray): The mask to be displayed.
        ax (matplotlib.axes.Axes): The axis on which to display the mask.
        obj_id (int, optional): Object ID for color selection. Defaults to None.
        random_color (bool, optional): If True, use a random color. Defaults to False.
    Returns:
        None
    """
    if random_color:
        color = np.concatenate([np.random.random(3), np.array([0.6])], axis=0)
    else:
        cmap = plt.get_cmap("tab10")
        cmap_idx = 0 if obj_id is None else obj_id
        color = np.array([*cmap(cmap_idx)[:3], 0.6])
    h, w = mask.shape[-2:]
    mask_image = mask.reshape(h, w, 1) * color.reshape(1, 1, -1)
    ax.imshow(mask_image)

def show_mask_image(mask, ax, random_color=False, borders = True):
    if random_color:
        color = np.concatenate([np.random.random(3), np.array([0.6])], axis=0)
    else:
        color = np.array([30/255, 144/255, 255/255, 0.6])
    h, w = mask.shape[-2:]
    mask = mask.astype(np.uint8)
    mask_image =  mask.reshape(h, w, 1) * color.reshape(1, 1, -1)
    if borders:
        import cv2
        contours, _ = cv2.findContours(mask,cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE) 
        # Try to smooth contours
        contours = [cv2.approxPolyDP(contour, epsilon=0.01, closed=True) for contour in contours]
        mask_image = cv2.drawContours(mask_image, contours, -1, (1, 1, 1, 0.5), thickness=2) 
    ax.imshow(mask_image)

def show_masks_image(image, masks, scores, point_coords=None, box_coords=None, input_labels=None, borders=True):
    """
    Display masks on a given image.
    Args:
        image (np.ndarray): The image on which to display the masks.
        masks (list): List of masks to be displayed.
        scores (list): List of scores corresponding to the masks.
        point_coords (list, optional): List of point coordinates. Defaults to None.
        box_coords (list, optional): List of box coordinates. Defaults to None.
        input_labels (list, optional): List of input labels. Defaults to None.
        borders (bool, optional): If True, display the borders of the masks. Defaults to True.
    Returns:
        None
    """
    for i, (mask, score) in enumerate(zip(masks, scores)):
        plt.figure(figsize=(10, 10))
        plt.imshow(image)
        show_mask_image(mask, plt.gca(), borders=borders)
        if point_coords is not None:
            assert input_labels is not None
            show_points(point_coords, input_labels, plt.gca())
        if box_coords is not None:
            # boxes
            show_box(box_coords, plt.gca())
        if len(scores) > 1:
            plt.title(f"Mask {i+1}, Score: {score:.3f}", fontsize=18)
        plt.axis('off')
        plt.show()

def show_points(coords, labels, ax, marker_size=200):
    """
    Display points on a given axis.
    Args:
        coords (np.ndarray): Array of point coordinates.
        labels (np.ndarray): Array of point labels (1 for positive, 0 for negative).
        ax (matplotlib.axes.Axes): The axis on which to display the points.
        marker_size (int, optional): Size of the markers. Defaults to 200.
    Returns:
        None
    """
    pos_points = coords[labels==1]
    neg_points = coords[labels==0]
    ax.scatter(pos_points[:, 0], pos_points[:, 1], color='green', marker='*', s=marker_size, edgecolor='white', linewidth=1.25)
    ax.scatter(neg_points[:, 0], neg_points[:, 1], color='red', marker='*', s=marker_size, edgecolor='white', linewidth=1.25)

def show_box(box, ax):
    """
    Display a bounding box on a given axis.
    Args:
        box (list or np.ndarray): The bounding box coordinates [x0, y0, x1, y1].
        ax (matplotlib.axes.Axes): The axis on which to display the box.
    Returns:
        None
    """
    x0, y0 = box[0], box[1]
    w, h = box[2] - box[0], box[3] - box[1]
    ax.add_patch(plt.Rectangle((x0, y0), w, h, edgecolor='green', facecolor=(0, 0, 0, 0), lw=2))

class video_predictor:
    """
    Segmentation anything 2 video predictor
    Args:
        model_cfg (str): Path to the model config file.
        sam2_checkpoint (str): Path to the model checkpoint file.
        device (str, optional): Device on which to run the model. Defaults to 'cpu'.
    Returns:
        None
    """
    def __init__(self,model_cfg,sam2_checkpoint,device='cpu') -> None:
        self.predictor = build_sam2_video_predictor(model_cfg, sam2_checkpoint, device=device)

    def inference_state(self,video_image_folder):
        """
        Initializing video from path of video frames folder
        """
        self.video_image_folder=video_image_folder
        self.frame_names = [f for f in listdir(self.video_image_folder) if isfile(join(self.video_image_folder, f))]
        self.frame_names = sorted(self.frame_names)
        self.joined_frame_names = [self.video_image_folder+'/'+self.frame_names[i] for i in range(len(self.frame_names))]
        self.inference_state = self.predictor.init_state(video_path=video_image_folder)

    def reset_state(self):
        """
        Reseting state to base
        """
        self.predictor.reset_state(self.inference)
    
    def predict_item(self,bbox:list[list]= None,points: list[list]=None,labels: list=None,frame_idx=0,show=True,image_loc=None,gemini_bbox=True,**kwargs):
        """
        Getting prediction from bounding box. Can add points to determine the location of segmentation
        """
        ann_obj_id=1
        if points and labels:
            points = np.array(points,dtype=np.float32)
            labels = np.array(labels,np.int32)
            print(f"points : {points}")
            print(f"labels : {labels}")
            prompts={}
            prompts[ann_obj_id] = points, labels

        if bbox:
            bbox = np.array(bbox,dtype=np.float32)
            if gemini_bbox:
                bbox = bbox[[1,0,3,2]]
            print(f"bbox : {bbox}")


        _, out_obj_ids, out_mask_logits = self.predictor.add_new_points_or_box(
                                            inference_state=self.inference_state,
                                            frame_idx=frame_idx,
                                            obj_id=ann_obj_id,
                                            points=points,
                                            labels=labels,
                                            box=bbox,**kwargs)

        if show:
            plt.figure(figsize=(9, 6))
            plt.title(f"frame {frame_idx}")
            if image_loc==None:
                image_loc =self.joined_frame_names[0]
            plt.imshow(Image.open(image_loc))
            try: 
                show_points(points, labels, plt.gca())
            except:
                pass
            for i, out_obj_id in enumerate(out_obj_ids):
                try:
                    show_points(*prompts[out_obj_id], plt.gca())
                except:
                    pass
                show_mask((out_mask_logits[i] > 0.0).cpu().numpy(), plt.gca(), obj_id=out_obj_id)

    def predict_video(self,vis_frame_stride=30):
        """
        Using the predicting item and propogate it through the all frames.
        """
        video_segments = {} 
        for out_frame_idx, out_obj_ids, out_mask_logits in self.predictor.propagate_in_video(self.inference_state):
            video_segments[out_frame_idx] = {out_obj_id: (out_mask_logits[i] > 0.0).cpu().numpy() for i, out_obj_id in enumerate(out_obj_ids)}

        plt.close("all")
        for out_frame_idx in range(0, len(self.frame_names), vis_frame_stride):
            plt.figure(figsize=(6, 4))
            plt.title(f"frame {out_frame_idx}")
            plt.imshow(Image.open(self.joined_frame_names[out_frame_idx]))
            for out_obj_id, out_mask in video_segments[out_frame_idx].items():
                show_mask(out_mask, plt.gca(), obj_id=out_obj_id)

class image_predictor:
    """
    Class for image prediction
    Args:
        model_cfg (str): Path to the model config file.
        sam2_checkpoint (str): Path to the model checkpoint file.
        device (str, optional): Device on which to run the model. Defaults to 'cpu'.
    Returns:
        None
    """
    def __init__(self,model_cfg,sam2_checkpoint,device='cpu') -> None:
        self.predictor = SAM2ImagePredictor(build_sam2(model_cfg, sam2_checkpoint, device=device))

    def set_image(self,image):
        """
        Setting image
        Args:
            image (str): Path to the image or numpy array
        Returns:
            None
        """
        if isinstance(image, str):
            image = cv2.imread(image)
            self.image = cv2.cvtColor(image,cv2.COLOR_BGR2RGB)
        else:
            self.image = image
        self.image_set = self.predictor.set_image(self.image)

    def predict_item(self,bbox:list[list]= None,points: list[list]=None,labels: list=None,show=True,gemini_bbox=True,**kwargs):
        """
        Getting prediction from bounding box or points.
        Args:
            bbox (list[list], optional): Bounding box. Defaults to None.
            points (list[list], optional): Points. Defaults to None.
            labels (list, optional): Labels. Defaults to None.
            show (bool, optional): Show the image. Defaults to True.
        Returns:
            masks, scores, logits
        """
        if points and labels:
            points = np.array(points,dtype=np.float32)
            labels = np.array(labels,np.int32)
            print(f"points : {points}")
            print(f"labels : {labels}")

        if bbox:
            bbox = np.array(bbox,dtype=np.float32)
            if gemini_bbox:
                bbox = bbox[[1,0,3,2]]
            print(f"bbox : {bbox}")

        predict_args ={}

        if points is not None:
            predict_args["point_coords"] = points
            predict_args["point_labels"] = labels

        if bbox is not None:
            predict_args["box"] = bbox
        print(predict_args)

        masks, scores, logits = self.predictor.predict(**predict_args,multimask_output=False,**kwargs)

        sorted_ind = np.argsort(scores)[::-1]
        masks = masks[sorted_ind]
        scores = scores[sorted_ind]
        logits = logits[sorted_ind]

        if show:
            plt.figure(figsize=(9, 6))
            plt.imshow(self.image)
            if points is not None:
                show_points(points, labels, plt.gca())
            if bbox is not None:
                show_box(bbox, plt.gca())
            # for i, mask in enumerate(masks):
            #     show_mask(mask, plt.gca(), random_color=True)
            show_masks_image(self.image, masks, scores, point_coords=points, box_coords=bbox, input_labels=labels)

        return masks, scores, logits
    
    
    



